/* Loads a file into memory and sends it to each client that connects.
   Efficiently scales to multiple clients at a time.
   Demonstrates the use of DDispatcher and DProtocol to create complex messages.
   Uses IPv6.
*/

#include "Dasyne.hpp"

using namespace std;
using namespace libdasyne;
class Client;

//The file to send to each client, contains the file size and a checksum
static DBuffer fileMessage;

const uint16_t PORT_NUMBER = 7890;//The port number to communicate over
static bool runProgram = true; //false if we should stop the server
static vector<Client*> theClients;//a list of clients who've connected

static DEngine netEngine;//calls event handlers for networking classes
static DServer *myServer = new DServer(&netEngine);//listens for client connections

//Loads a file with the specified name into memory, and populates fileMessage
bool loadFile(string fileName);

//_______________ Client class, one per connection _______________

class Client : public DDispatcher
{
    public:
        //constructor, passes the DEngine and an existing connection to DClient
        Client(boost::asio::ip::tcp::socket *newSocket) : DDispatcher(&netEngine, newSocket)
        {
            setErrorHandler(boost::bind(&Client::errorHandler, this, _1, _2));
            setSendHandler(boost::bind(&Client::sendHandler, this));
        }//constructor

    private:
        //Event handler called when a network error occurs
        void errorHandler(string message,
                          bool socketOK)
        {
            cout << message << endl;
        }//errorHandler

        //Event handler, called when we've sent all pending data
        void sendHandler(void)
        {
            cout << "Sent all data to " << getRemoteAddress() << endl;
        }//sendHandler
};//Client

//_______________ DServer Event Handlers _______________

//Called when a client connects
void connectHandler(boost::asio::ip::tcp::socket *newSocket)
{
    Client *newClient = new Client(newSocket);

    cout << "A client connected from " << newClient->getRemoteAddress() << endl;
    theClients.push_back(newClient);
    newClient->sendMessage(fileMessage);
}//connectHandler

//Called when there's an error accepting connections
void errorHandler(string errMsg)
{
    cout << errMsg << endl;
    runProgram = false;
}//errorHandler

//_______________ Main Function _______________

int main(int argc,
         char *argv[])
{
    string userText;//text entered by the user
    int eventsRun = 0;//the number of events dispatched by a call to update()

    //### Verify that a file name was given ###
    if (argc != 2)
    {
        cout << "Expected a file name" << endl;
        runProgram = false;
    }//if argc

    //### Load the file into memory ###
    if (runProgram)
    {
        if (!loadFile(argv[1]))
            runProgram = false;
    }//if runProgram

    //### Bind the listen socket ###
    if (runProgram)
    {
        if (!myServer->bind(libdasyne::DA_IPV6, PORT_NUMBER))
        {
            runProgram = false;
            cout << "Couldn't bind: " << myServer->getErrMsg() << endl;
        }//if !bind()
    }//if runProgram

    //### Start listening for connections ###
    if (runProgram)
    {
        myServer->setErrorHandler(&errorHandler);
        if (myServer->startListening(&connectHandler))
            cout << "Listening for TCP connections, hit Ctrl-C to exit." << endl;
        else
        {
            runProgram = false;
            cout << myServer->getErrMsg() << endl;
        }//else startListening()
    }//if runProgram

    //### Periodically call event handlers ###
    while (runProgram)
    {
        eventsRun = netEngine.update(-1);
        if (eventsRun < 0)
        {
            cout << netEngine.getErrMsg();
            runProgram = false;
        }//if eventsRun

        //Sleep if we have nothing to do
        else if (eventsRun == 0)
            DEngine::sleep(1);
    }//while runProgram

    //### Delete the objects that use DEngine before it's deleted ###
    for (unsigned int i = 0; i < theClients.size(); i++)
        delete theClients.at(i);

    delete myServer;

    return 0;
}//main

bool loadFile(string fileName)
{
    DField *fileField = NULL;//the current field to add to our fileMessage
    DProtocol fileProtocol;//constructs the message to send to each client
    DBuffer theFile;//the entire file, loaded from disk
    ifstream infile;//obtains binary input from the file
    uint32_t checksum = 0;//a CRC32 checksum of the file contents
    uint8_t oneByte = 0;//a single byte from the file
    bool result = true;//false if we couldn't load the file.  Either file not found,
                      //or we ran out of memory

    //### Open the file ###
    infile.open(fileName.c_str(), ios::binary);
    if (!infile.is_open())
    {
        result = false;
        cout << "Couldn't open " << fileName << endl;
    }//if !is_open()

    //### Load the file ###
    if (result)
    {
        cout << "Loading " << fileName << "... ";

        oneByte = infile.get();
        while (!infile.eof())
        {
            theFile.appendByte(oneByte);
            oneByte = infile.get();
        }//while !eof()

        cout << "complete." << endl;
        cout << "File size is " << theFile.size() << " bytes" << endl;

        checksum = theFile.crc32();
        cout << "Checksum: " << checksum << endl;
        cout << endl;
    }//if result

    //### Create the message ###
    if (result)
    {
        fileField = new DField("checksum", DA_UINT32_T);
        fileField->addValue(checksum);
        fileProtocol.appendField(fileField);

        fileField = new DField("file contents", DA_DBUFFER);
        fileField->addDBuffer(theFile);
        fileProtocol.appendField(fileField);

        fileField = new DField("file name", DA_STRING);
        fileField->addString(fileName);
        fileProtocol.appendField(fileField);

        fileMessage = fileProtocol.generateMessage();
    }//if result

    //### Close the file ###
    if (infile.is_open())
        infile.close();

    return result;
}//loadFile
