/* Checks the functionality of the DBuffer class, used for internal debugging.
*/

#include <iostream>
#include "Dasyne.hpp"

using namespace std;
using namespace libdasyne;

//Test appending and removing elements from a DBytes
void testDBytes(void);

//Test iterating through a DBytes
void testDBufferIter(void);

//Test compression and string conversions
void testDBuffer(void);

int main(int argc,
         char *argv[])
{
    cout << "Array Size: " << DBytes::ARRAY_SIZE << endl;

    cout << endl;
    cout << "############### Testing DBytes ###############" << endl;
    cout << endl;
    testDBytes();

    cout << endl;
    cout << "############### Testing DBufferIter ###############" << endl;
    cout << endl;
    testDBufferIter();

    cout << endl;
    cout << "############### Testing DBuffer ###############" << endl;
    cout << endl;
    testDBuffer();
}//main

void testDBuffer(void)
{
    DBuffer buffer;
    float testFloat = 666;
    double testDouble = 6677143;
    uint8_t testByte = 32;
    unsigned int lastIndex = 0;//the index of the last element in buffer

    cout << "Appending Hello World!" << endl;
    buffer.appendString("Hello World!");
    cout << buffer.toString() << endl;

    cout << endl;
    cout << "Appending  Goodbye Universe!" << endl;
    buffer.appendString(" Goodbye Universe!");
    cout << buffer.toString() << endl;

    for (unsigned int i = 0; i < 10; i++)
        buffer.appendString(" I am compressable data!");

    cout << endl;
    cout << "Data size: " << buffer.size() << endl;
    DBuffer compressed = buffer.compress(9);
    if (compressed.size() == 0)
    {
        cout << buffer.getErrMsg() << endl;
        throw 1;
    }//if size()

    cout << "Compressed size: " << compressed.size() << endl;

    DBuffer decompressed = compressed.decompress();
    if (decompressed.size() == 0)
    {
        cout << buffer.getErrMsg() << endl;
        throw 1;
    }//if size()

    cout << "Decompressed size: " << decompressed.size() << endl;
    cout << "Decompressed: " << decompressed.toString() << endl;

    cout << endl;
    cout << "Compressing self..." << endl;
    buffer = buffer.compress(9);

    cout << endl;
    buffer.clear();
    cout << "Appending float: " << buffer.appendVariable(testFloat) << " bytes, " << testFloat << endl;
    cout << "Inserting double: " << buffer.insertVariable(testDouble) << " bytes, " << testDouble << endl;
    cout << "Appending byte: " << buffer.appendVariable(testByte) << " bytes, " << static_cast<int>(testByte) << endl;

    cout << endl;
    cout << "CRC 32: " << buffer.crc32() << endl;

    cout << endl;
    testFloat = 0;
    testDouble = 0;
    testByte = 0;
    lastIndex = buffer.size() - 1;

    buffer.extractVariable(0, testDouble);
    buffer.extractVariable(lastIndex, testByte);
    buffer.extractVariable(lastIndex - sizeof(float), testFloat);

    cout << "Extracted: " << testFloat << ", " << testDouble << ", " << static_cast<int>(testByte) << endl;
    if ((testFloat != 666) || (testDouble != 6677143) || (testByte != 32))
    {
        cout << "Error converting to/from bytes" << endl;
        throw 1;
    }//if testFloat ... || testByte
}//testDBuffer

void testDBufferIter(void)
{
    DTime iterTime;
    DBufferIter dIter;
    DBytes buffer;
    volatile uint8_t oneByte = 0;
    unsigned long int bufferSize = 0;

    cout << "Appending data" << endl;
    for (unsigned int i = 0; i < 25; i++)
    {
        buffer.appendByte(i);
        cout << i << " ";
    }//for i
    cout << endl;

    dIter.setBuffer(&buffer);
    dIter.setIndex(0);

    while (!dIter.finished())
    {
        cout << static_cast<int>(dIter.value()) << " ";
        dIter.next();
    }//while !finished()
    cout << endl;

    cout << endl;
    cout << "Removing 1 byte from front" << endl;
    buffer.removeFirst(1);

    dIter.setIndex(0);

    while (!dIter.finished())
    {
        cout << static_cast<int>(dIter.value()) << " ";
        dIter.next();
    }//while !finished()

    cout << endl;

    cout << endl;
    cout << "Multiplying by 2" << endl;

    dIter.setIndex(0);

    while (!dIter.finished())
    {
        dIter.value() = dIter.value() * 2;
        cout << static_cast<int>(dIter.value()) << " ";
        dIter.next();
    }//while !finished()

    cout << endl;

    cout << endl;
    cout << "Adding 1 MB of data" << endl;
    buffer.clear();
    for (unsigned int i = 0; i < 1000000; i++)
        buffer.appendByte(i % 255);

    cout << "Iterating through it 100 times with DBufferIter" << endl;
    iterTime.updateTime();

    for (unsigned int i = 0; i < 100; i++)
    {
        dIter.setIndex(0);

        while (!dIter.finished())
        {
            oneByte = dIter.value();
            dIter.next();
        }//while !finished()
    }//for i

    cout << "Time: " << iterTime.timeElapsed() << endl;
    cout << endl;

    cout << "Iterating through it 100 times with getByte()" << endl;
    iterTime.updateTime();
    bufferSize = buffer.size();

    for (unsigned int i = 0; i < 100; i++)
    {
        for (unsigned int j = 0; j < bufferSize; j++)
            oneByte = buffer.getByte(j);
    }//for i

    cout << "Time: " << iterTime.timeElapsed() << endl;
    cout << endl;
}//testDBufferIter

void testDBytes(void)
{
    DBytes first;

    cout << "Appending data" << endl;
    for (unsigned int i = 0; i < 25; i++)
    {
        first.appendByte(i);
        cout << i << " ";
    }//for i
    cout << endl;

    for (unsigned int i = 0; i < first.size(); i++)
        cout << static_cast<int>(first.getByte(i)) << " ";
    cout << endl;

    cout << "Data size: " << first.size() << endl;

    cout << endl;
    cout << "Modifying data" << endl;
    for (unsigned int i = 0; i < first.size(); i++)
    {
        first.setByte(i, i * 2);
        cout << i * 2 << " ";
    }//for i
    cout << endl;

    for (unsigned int i = 0; i < first.size(); i++)
        cout << static_cast<int>(first.getByte(i)) << " ";
    cout << endl;

    cout << endl;
    cout << "Inserting 12 bytes to front" << endl;
    for (unsigned int i = 0; i < 12; i++)
        first.insertByte(i + 100);

    for (unsigned int i = 0; i < first.size(); i++)
        cout << static_cast<int>(first.getByte(i)) << " ";
    cout << endl;

    cout << endl;
    cout << "Size: " << first.size() << endl;
    cout << "Removing 15 bytes from front" << endl;
    first.removeFirst(15);

    for (unsigned int i = 0; i < first.size(); i++)
        cout << static_cast<int>(first.getByte(i)) << " ";
    cout << endl;
    cout << "New Size: " << first.size() << endl;

    cout << endl;
    cout << "Removing 1 byte from front" << endl;
    first.removeFirst(1);

    for (unsigned int i = 0; i < first.size(); i++)
        cout << static_cast<int>(first.getByte(i)) << " ";
    cout << endl;

    cout << endl;
    cout << "Removing 1 byte from front" << endl;
    first.removeFirst(1);

    for (unsigned int i = 0; i < first.size(); i++)
        cout << static_cast<int>(first.getByte(i)) << " ";
    cout << endl;

    cout << endl;
    cout << "Removing 17 bytes from front" << endl;
    first.removeFirst(17);

    for (unsigned int i = 0; i < first.size(); i++)
        cout << static_cast<int>(first.getByte(i)) << " ";
    cout << endl;

    cout << endl;
    cout << "Removing 1 byte from front" << endl;
    first.removeFirst(1);

    for (unsigned int i = 0; i < first.size(); i++)
        cout << static_cast<int>(first.getByte(i)) << " ";
    cout << endl;

    cout << endl;
    cout << "Appending 6 bytes" << endl;
    uint8_t byteArray[6];

    for (unsigned int i = 0; i < 6; i++)
    {
        byteArray[i] = i;
        cout << i << " ";
    }//for i
    cout << endl;
    first.appendArray(byteArray, 6);


    for (unsigned int i = 0; i < first.size(); i++)
        cout << static_cast<int>(first.getByte(i)) << " ";
    cout << endl;

    cout << "Data size: " << first.size() << endl;

    DBytes second = first;

    cout << endl;
    cout << "Copy: " << endl;
    for (unsigned int i = 0; i < second.size(); i++)
        cout << static_cast<int>(second.getByte(i)) << " ";
    cout << endl;

    if (second.size() != first.size())
    {
        cout << "ERROR: DBytes result has an invalid size" << endl;
        cout << "It's " << second.size() << ", should be " << first.size() << endl;
        throw 1;
    }//if size()

    if ((second.getByte(0) != 46) ||
        (second.getByte(1) != 48) ||
        (second.getByte(2) != 0) ||
        (second.getByte(3) != 1) ||
        (second.getByte(4) != 2) ||
        (second.getByte(5) != 3) ||
        (second.getByte(6) != 4) ||
        (second.getByte(7) != 5))
    {
        cout << "ERROR: DBytes result has an invalid value" << endl;
        throw 1;
    }//if getByte() ... || getByte()
}//testDBytes
