/*
 * Copyright (C) 2005-2011 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

/* DESCRIPTION
    The DBuffer class adds features to the DBytes class, which it inherits from.
    These include compression, hashing, and conversion to and from strings. It also
    lets us append binary data from variables. For example, appending a float adds four
    bytes to the DBuffer. All data is converted to network-order endianness. It can then
    be retrieved with extractVariable().

    These features are in a separate class from DBytes to restrict the number of places
    where DBytes' private member variables can be accessed, for ease in debugging.
*/

#ifndef DBUFFER_HPP_INCLUDED
#define DBUFFER_HPP_INCLUDED

#include "DBytes.hpp"
#include "DEngine.hpp"

namespace libdasyne
{
    class DBufferIter;

    class DBuffer : public DBytes
    {
        public:
            //constructor
            DBuffer(void);

            //copy constructor
            DBuffer(const DBuffer &source);

            //destructor
            ~DBuffer(void);

            //Overloaded = operator
            DBuffer& operator= (const DBuffer &source);

            //returns the string version of the data (corrupts binary data)
            std::string toString(void) const;

            //Appends source to the end of the data (corrupts binary data)
            void appendString(const std::string &source);

            //Return a description of any errors that occur
            std::string getErrMsg(void) const;

            //Returns a compressed version of this DBuffer. compressionLevel should be between 0
            //and 9. Returns an empty DBuffer if errors occur, and sets errMsg appropriately.
            DBuffer compress(int compressionLevel);

            //decompresses and returns the data in this object.  Returns an empty DBuffer if errors occur,
            //and sets errMsg.
            DBuffer decompress(void);

            //calculate a 32-bit CRC for theData
            uint32_t crc32(void) const;

            //Appends a binary variable to the end of the DBytes. Returns the number of bytes
            //appended (the size of theVariable).
            template <class T>
            unsigned int appendVariable(T theVariable);

            //Inserts a binary variable to the front of the DBytes. Returns the number of bytes
            //inserted (the size of theVariable).
            template <class T>
            unsigned int insertVariable(T theVariable);

            //Extracts a binary variable from the specified index. Returns false if the index is
            //out of bounds
            template <class T>
            void extractVariable(long int index,
                                 T &theVariable) const;

            //the number of bytes to send/receive to zlib at a time
            //should be between 1 and 32767 inclusive
            static const int ZLIB_BUFFER_SIZE = 1024;

        private:
            //used by the fast crc32 algorithm
            static uint32_t *crcTable;

            //contains a description of any errors that occur
            std::string errMsg;

            //############### FUNCTIONS ###############

            //initialize crcTable for the fast crc32 algorithm
            static uint32_t* crcInit(void);

            //miscellaneous method needed by the crc32 algorithm
            uint32_t reflect(uint32_t data,
                             uint8_t nBits) const;

            //1) get the deflation stream ready to compress
            bool initializeDeflation(z_stream &defStream,
                                     int compressionLevel);

            //2) load the data into zlib
            bool runDeflation(z_stream& defStream,
                              DBuffer &compressedBytes,
                              uint8_t outputBuffer[]);

            //3) get the compressed data from zlib
            bool finishDeflation(z_stream& defStream,
                                 DBuffer &compressedBytes,
                                 uint8_t outputBuffer[]);

            //4) close the z_stream
            bool endDeflation(z_stream& defStream);

            //fill inputBuffer with bytes
            int fillCompressionInput(uint8_t inputBuffer[],
                                     DBufferIter &dIter);

            //store the compressed data from zlib
            void emptyCompressionOutput(uint8_t outputBuffer[],
                                        int bufferLength,
                                        DBuffer &compressedBytes);

            //1) get the inflation stream ready to decompress
            bool initializeInflation(z_stream &infStream);

            //2) get the uncompressed data from zlib
            bool runInflation(z_stream& infStream,
                              DBuffer &extractedBytes);

    };//DBuffer
}//libdasyne

template <class T>
void libdasyne::DBuffer::extractVariable(long int index,
                                         T &theVariable) const
{
    libdasyne::EndianType endianness = DEngine::getEndianness();
    unsigned int bytePos = index;//position in theData[] to read from

    if ((index < 0) || (index + sizeof(theVariable) > size()))
        DEngine::fatalError("DBuffer::extractVariable(): out of bounds");

    if (endianness == libdasyne::DA_LITTLE_ENDIAN)
    {
        for (int i = sizeof(theVariable) - 1; i >= 0; i--)
        {
            ((uint8_t*)&theVariable)[i] = getByte(bytePos);
            bytePos++;
        }//for i
    }//if getEndianness

    else if (endianness == libdasyne::DA_BIG_ENDIAN)
    {
        for (unsigned int i = 0; i < sizeof(theVariable); i++)
        {
            ((uint8_t*)&theVariable)[i] = getByte(bytePos);
            bytePos++;
        }//for i
    }//else if endianness

    else
        DEngine::fatalError("DBuffer::extractVariable(): unknown endianness");
}//extractVariable

template <class T>
unsigned int libdasyne::DBuffer::insertVariable(T theVariable)
{
    libdasyne::EndianType endianness = DEngine::getEndianness();
    uint8_t oneByte = ' ';//a single byte from theVariable

    if (endianness == libdasyne::DA_LITTLE_ENDIAN)
    {
        for (unsigned int i = 0; i < sizeof(theVariable); i++)
        {
            oneByte = ((uint8_t*)&theVariable)[i];
            insertByte(oneByte);
        }//for i
    }//if endianness

    else if (endianness == libdasyne::DA_BIG_ENDIAN)
    {
        for (int i = sizeof(theVariable) - 1; i >= 0; i--)
        {
            oneByte = ((uint8_t*)&theVariable)[i];
            insertByte(oneByte);
        }//for i
    }//else if endianness

    else
        DEngine::fatalError("DBuffer::insertVariable(): unknown endianness");

    return sizeof(theVariable);
}//insertVariable

template <class T>
unsigned int libdasyne::DBuffer::appendVariable(T theVariable)
{
    libdasyne::EndianType endianness = DEngine::getEndianness();
    uint8_t oneByte = ' ';//a single byte from theVariable

    if (endianness == libdasyne::DA_LITTLE_ENDIAN)
    {
        for (int i = sizeof(theVariable) - 1; i >= 0; i--)
        {
            oneByte = ((uint8_t*)&theVariable)[i];
            appendByte(oneByte);
        }//for i
    }//if endianness

    else if (endianness == libdasyne::DA_BIG_ENDIAN)
    {
        for (unsigned int i = 0; i < sizeof(theVariable); i++)
        {
            oneByte = ((uint8_t*)&theVariable)[i];
            appendByte(oneByte);
        }//for i
    }//else if endianness

    else
        DEngine::fatalError("DBytes::appendVariable(): unknown endianness");

    return sizeof(theVariable);
}//appendVariable

#endif
