/*
 * Copyright (C) 2011 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

/* DESCRIPTION
    The DEvent class contains an update function that will be called whenever the associated
    DEngine updates. This allows us to implement custom events, by inheriting from DEvent
    and re-implementing the update function. The resolution of the events will depend on how
    often DEngine::update() is called.
*/

#ifndef DEVENT_HPP_INCLUDED
#define DEVENT_HPP_INCLUDED

#include <iostream>
#include <list>

namespace libdasyne
{
    class DEngine;

    class DEvent
    {
        public:
            //constructor
            DEvent(DEngine *newEngine);

            //copy constructor
            DEvent(const DEvent &source);

            //destructor
            virtual ~DEvent(void);

            //Overloaded = operator
            DEvent& operator= (const DEvent &source);

            //Event handler, called when the DEngine updates. Returns the number of events
            //dispatched, this should normally be one.
            virtual unsigned int update(void) = 0;

            //Report that the DEngine has been deleted
            void reportEngineLost(DEngine *theEngine);

        protected:
            //Causes this event to leave the current engine (if any), and register itself
            //with newEngine
            void joinEngine(DEngine *newEngine);

        private:
            //Iterator to my location in the engine's event list. Only valid if engine != NULL
            std::list<DEvent*>::iterator myIter;

            //Pointer to the DEngine that's supplying our events.
            DEngine *engine;

            //De-registers with the current engine, if any.
            void leaveEngine(void);

    };//DEvent
}//namespace libdasyne

#endif
