/*
 * Copyright (C) 2011 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

/* DESCRIPTION
  An event-based timer. It will periodically call elapsed() at the specified interval.
  The events are triggered whenever the DEngine associated with this timer updates, and so
  the timer resolution is dependent on how often DEngine::update is called.
*/

#ifndef DTIMER_HPP_INCLUDED
#define DTIMER_HPP_INCLUDED

#include "DEvent.hpp"
#include "DEngine.hpp"

namespace libdasyne
{
    class DTime;
    class DTimerEvent;

    class DTimer
    {
        public:
            //constructor
            DTimer(DEngine *newEngine);

            //copy constructor
            DTimer(const DTimer &source);

            //destructor
            virtual ~DTimer(void);

            //Overloaded = operator
            DTimer& operator= (const DTimer &source);

            //Set the timer interval in milliseconds
            void setInterval(unsigned int newInterval);

            //Accessors
            unsigned int getInterval(void);

            //Returns true if this timer is currently running
            bool isRunning(void);

            //Starts the timer, it will call the event handler at the specified interval
            void start(void);

            //Stops the timer
            void stop(void);

            //Called when the DEngine updates. We'll check the current time, and decide whether
            //or not to call elapsed()
            unsigned int update(void);

            //Specify a function to call when the timer expires. It should have a signature like:
            //void foo(void).
            void setHandler(boost::function<void (void)> newHandler);

        private:
            //Tells us when we should fire our timeout() event
            DTime *myTime;

            //The engine to use for event-handling
            DEngine *engine;

            //Registers with a DEngine, calls update when the DEngine updates
            DTimerEvent *eventHandler;

            //Function to call when the timer elapses
            boost::function<void (void)> handler;

            //The interval between calls to elapsed(), in milliseconds
            unsigned int interval;

            //True if the timer is running, false otherwise
            bool running;

            //True if we should call the handler function pointer instead of
            //elapsed() when the timer expires
            bool haveHandler;

            //Copies the member variables from source.
            void copyVariables(const DTimer &source);

    };//class DTimer
}//namespace libdasyne
#endif
